import utool as ut
import ubelt as ub
try:
    import guitool_ibeis as gt
    from guitool_ibeis import mpl_widget
    INSPECT_BASE = gt.GuitoolWidget
    MatplotlibWidget = mpl_widget.MatplotlibWidget
except ImportError:
    import warnings
    warnings.warn('WARNING: guitool not available')
    MatplotlibWidget = object
    INSPECT_BASE = object


def lazy_test_annot(key):
    import numpy as np
    rchip_fpath = ut.grab_test_imgpath(key)
    annot = ut.LazyDict({
        'aid': key.split('.')[0],
        'nid': key[0:4],
        'rchip_fpath': rchip_fpath,
        'gps': (np.nan, np.nan),
        'yaw': np.nan,
        'view': np.nan,
        'qual': np.nan,
        'time': np.nan,
    })
    return annot

try:
    import dtool_ibeis as dt

    MatchDisplayConfig = dt.from_param_info_list([
        ut.ParamInfo('overlay', True),
        ut.ParamInfo('show_all_kpts', False),
        ut.ParamInfo('mask_blend', 0.0, min_=0, max_=1),

        ut.ParamInfo('heatmask', True, hideif=':not overlay'),
        ut.ParamInfo('show_homog', False, hideif=':not overlay'),
        ut.ParamInfo('show_ori', False, hideif=':not overlay'),
        ut.ParamInfo('show_ell', False, hideif=':not overlay'),
        ut.ParamInfo('show_pts', False, hideif=':not overlay'),
        ut.ParamInfo('show_lines', False, hideif=lambda cfg: not cfg['overlay']),
        ut.ParamInfo('show_rect', False, hideif=':not overlay'),
        ut.ParamInfo('show_eig', False, hideif=':not overlay'),
        ut.ParamInfo('ell_alpha', 0.6, min_=0, max_=1, hideif=':not overlay'),
        ut.ParamInfo('line_alpha', 0.35, min_=0, max_=1, hideif=':not overlay'),
    ])
except ImportError:
    pass


class MatchInspector(INSPECT_BASE):
    """
    A widget that contains
        (1) a viewport that displays an annotation pair with matches overlayed.
        (2) a control panel for tuning matching parameters
        (3) a text area displaying information about the match vector

    CommandLine:
        python -m vtool_ibeis.inspect_matches MatchInspector:0 --show
        python -m vtool_ibeis.inspect_matches MatchInspector:1 --show

        python -m vtool_ibeis.inspect_matches MatchInspector:1 --db GZ_Master1 --aids=1041,1045 --show

    Example:
        >>> # SCRIPT
        >>> from vtool_ibeis.inspect_matches import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> gt.ensure_qapp()
        >>> ut.qtensure()
        >>> annot1 = lazy_test_annot('tsukuba_r')
        >>> annot2 = lazy_test_annot('tsukuba_l')
        >>> match = vt.PairwiseMatch(annot1, annot2)
        >>> self = MatchInspector(match=match)
        >>> self.show()
        >>> # xdoctest: +REQUIRES(--show)
        >>> #self.update()
        >>> gt.qtapp_loop(qwin=self, freq=10)

    Example:
        >>> # SCRIPT
        >>> from vtool_ibeis.inspect_matches import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> import ibeis
        >>> gt.ensure_qapp()
        >>> ut.qtensure()
        >>> ibs = ibeis.opendb(defaultdb='PZ_MTEST')
        >>> aids = ub.argval('--aids', default=[1, 2])
        >>> print('aids = %r' % (aids,))
        >>> annots = ibs.annots(aids)
        >>> annot1 = annots[0]._make_lazy_dict()
        >>> annot2 = annots[1]._make_lazy_dict()
        >>> cfgdict = MatchDisplayConfig().asdict()
        >>> cfgdict = ut.argparse_dict(cfgdict)
        >>> match = vt.PairwiseMatch(annot1, annot2)
        >>> self = MatchInspector(match=match, cfgdict=cfgdict)
        >>> self.show()
        >>> # xdoctest: +REQUIRES(--show)
        >>> #self.update()
        >>> gt.qtapp_loop(qwin=self, freq=10)
    """

    def showEvent(self, event):
        super(MatchInspector, self).showEvent(event)
        # Fire initialize event after we show the GUI
        # QtCore.QTimer.singleShot(50, self.init_inference)
        self.first_show()

    def first_show(self, state=None):
        if self.match is not None:
            # Show the match if updating is on, otherwise just draw the annot
            # pair
            if self.autoupdate_cb.checkState():
                self.update()
            else:
                self.draw_pair()

    def set_match(self, match=None, on_context=None, info_text=None):
        self.match = match
        self.info_text = info_text
        self.on_context = on_context
        if self.isVisible():
            self.first_show()

    def initialize(self, match=None, on_context=None, autoupdate=True,
                   info_text=None, cfgdict=None):
        from plottool_ibeis import abstract_interaction
        from guitool_ibeis.__PYQT__ import QtCore
        self.set_match(match, on_context, info_text)
        self._setup_configs(cfgdict=cfgdict)
        self._setup_layout(autoupdate=autoupdate)
        abstract_interaction.register_interaction(self)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.execContextMenu)

    def execContextMenu(self, qpoint):
        if self.on_context:
            options = self.on_context()
        else:
            options = [('No context set', None)]
        gt.popup_menu(self, qpoint, options)

    def screenshot(self):
        import plottool_ibeis as pt
        with pt.RenderingContext() as render:
            self.match.show(**self.disp_config)
        fpaths = gt.newFileDialog('.', mode='save', exec_=True)
        if fpaths is not None and len(fpaths) > 0:
            fpath = fpaths[0]
            if not fpath.endswith('.jpg'):
                fpath += '.jpg'
            import vtool_ibeis as vt
            vt.imwrite(fpath, render.image)

    def embed(self):
        match = self.match  # NOQA
        import utool
        utool.embed()

    def _new_config_widget(self, cfg, changed=None):
        from guitool_ibeis import PrefWidget2
        user_mode = 0
        cfg_widget = PrefWidget2.EditConfigWidget(
            config=cfg, user_mode=user_mode, parent=self, changed=changed)
        return cfg_widget

    def closeEvent(self, event):
        from plottool_ibeis import abstract_interaction
        abstract_interaction.unregister_interaction(self)
        super(MatchInspector, self).closeEvent(event)

    def _setup_configs(self, cfgdict=None):
        from vtool_ibeis import matching
        import dtool_ibeis
        # import pyhesaff

        # default_dict = pyhesaff.get_hesaff_default_params()
        # default_dict = vt.get_extract_features_default_params()
        TmpFeatConfig = dtool_ibeis.from_param_info_list(matching.VSONE_FEAT_CONFIG)

        TmpNChipConfig = dtool_ibeis.from_param_info_list(matching.NORM_CHIP_CONFIG)
        # [
        #     ut.ParamInfo(key, val) for key, val in default_dict.items()
        #     # ut.ParamInfo('affine_invariance', True),
        #     # ut.ParamInfo('rotation_invariance', False),
        # ])

        self.featconfig = TmpFeatConfig()
        self.chipconfig = TmpNChipConfig()

        TmpVsOneConfig = dtool_ibeis.from_param_info_list(
            matching.VSONE_DEFAULT_CONFIG)
        self.config = TmpVsOneConfig()
        self.disp_config = MatchDisplayConfig()

        if cfgdict is not None:
            print('[inspect_match] default cfgdict = %r' % (cfgdict,))
            self.config.update(**cfgdict)
            self.featconfig.update(**cfgdict)
            self.chipconfig.update(**cfgdict)
            self.disp_config.update(**cfgdict)

        # Make config widgets after setting defaults
        self.chipconfig_widget = self._new_config_widget(
            self.chipconfig, changed=self.on_chip_cfg_changed)
        self.featconfig_widget = self._new_config_widget(
            self.featconfig, changed=self.on_feat_cfg_changed)
        self.config_widget = self._new_config_widget(
            self.config, changed=self.on_cfg_changed)
        self.disp_config_widget = self._new_config_widget(
            self.disp_config, changed=self.on_cfg_changed)

    def _setup_layout(self, autoupdate=True):
        from guitool_ibeis.__PYQT__ import QtWidgets
        self.menubar = gt.newMenubar(self)
        self.menuFile = self.menubar.newMenu('Dev')
        self.menuFile.newAction(triggered=self.embed)
        self.menuFile.newAction(triggered=self.screenshot)
        splitter1 = self.addNewSplitter(orientation='horiz')
        config_vframe = splitter1.newWidget()
        splitter2     = splitter1.addNewSplitter(orientation='vert')
        config_vframe.addWidget(QtWidgets.QLabel('Chip Config'))
        config_vframe.addWidget(self.chipconfig_widget)
        config_vframe.addWidget(QtWidgets.QLabel('Feat Config'))
        config_vframe.addWidget(self.featconfig_widget)
        config_vframe.addWidget(QtWidgets.QLabel('Query Config'))
        config_vframe.addWidget(self.config_widget)
        config_vframe.addWidget(QtWidgets.QLabel('Display Config'))
        config_vframe.addWidget(self.disp_config_widget)
        # update_hframe = config_vframe.addNewWidget(orientation='horiz')
        # update_hframe.addNewButton('Update', pressed=self.update)
        self.autoupdate_cb = config_vframe.addNewCheckBox(
            'auto-update', checked=autoupdate, changed=self.first_show)

        self.mpl_widget = MatplotlibWidget(parent=self)
        splitter2.addWidget(self.mpl_widget)

        self.infobox = splitter2.addNewTextEdit()

    def execute_vsone(self):
        from vtool_ibeis import matching
        print('[inspect_match] Execute vsone')

        cfgdict = {}
        cfgdict.update(self.featconfig.asdict())
        cfgdict.update(self.chipconfig.asdict())

        match = self.match
        match.verbose = True
        match._inplace_default = True
        matching.ensure_metadata_vsone(match.annot1, match.annot2,
                                       cfgdict=cfgdict)

        match_config = self.config.asdict()
        match.apply_all(match_config)

    def draw_pair(self):
        if self.match is None:
            return
        self.mpl_widget.clf()
        ax = self.mpl_widget.ax
        info_html = ''
        if self.info_text is not None:
            info_html = '<pre>' + self.info_text + '</pre>'
        self.infobox.setText(info_html)
        self.match.show(ax=ax, overlay=False)
        self.mpl_widget.fig.canvas.draw()

    def draw_vsone(self):
        match = self.match
        summary = match._make_local_summary_feature_vector(summary_ops={'sum'})
        info_html = ''
        if self.info_text is not None:
            info_html = '<pre>' + self.info_text + '</pre>'
        feat_html = '<pre>' + ut.align(ub.repr2(summary), ':') + '</pre>'
        self.infobox.setText(info_html + feat_html)

        self.mpl_widget.clf()
        ax = self.mpl_widget.ax
        match.show(ax=ax, **self.disp_config)
        #fig.show()
        self.mpl_widget.fig.canvas.draw()

    def update(self, state=None):
        if self.autoupdate_cb.checkState() and self.match is not None:
            self.execute_vsone()
            self.draw_vsone()

    def on_cfg_changed(self, *args):
        self.update()
        self.cfg_needs_update = True

    def on_chip_cfg_changed(self, *args):
        print('Update feats')
        feat_keys = ['nchip', 'vecs', 'kpts', '_feats', 'flann']
        self.match.annot1._mutable = True
        self.match.annot2._mutable = True
        for key in feat_keys:
            if key in self.match.annot1:
                del self.match.annot1[key]
            if key in self.match.annot2:
                del self.match.annot2[key]
        self.update()
        self.cfg_needs_update = True

    def on_feat_cfg_changed(self, *args):
        print('Update feats')
        feat_keys = ['vecs', 'kpts', '_feats', 'flann']
        self.match.annot1._mutable = True
        self.match.annot2._mutable = True
        for key in feat_keys:
            if key in self.match.annot1:
                del self.match.annot1[key]
            if key in self.match.annot2:
                del self.match.annot2[key]
        self.update()
        self.cfg_needs_update = True


def make_match_interaction(matches, metadata, type_='RAT+SV', **kwargs):
    import plottool_ibeis.interact_matches
    #import plottool_ibeis as pt
    fm, fs = matches[type_][0:2]
    try:
        H1 = metadata['H_' + type_.split('+')[0]]
    except Exception:
        H1 = None
    #fm, fs = matches['RAT'][0:2]
    annot1 = metadata['annot1']
    annot2 = metadata['annot2']
    rchip1, kpts1, vecs1 = ub.dict_take(annot1, ['nchip', 'kpts', 'vecs'])
    rchip2, kpts2, vecs2 = ub.dict_take(annot2, ['nchip', 'kpts', 'vecs'])
    #pt.show_chipmatch2(rchip1, rchip2, kpts1, kpts2, fm=fm, fs=fs)
    fsv = fs[:, None]
    interact = plottool_ibeis.interact_matches.MatchInteraction2(
        rchip1, rchip2, kpts1, kpts2, fm, fs, fsv, vecs1, vecs2, H1=H1,
        **kwargs)
    return interact


def show_matching_dict(matches, metadata, *args, **kwargs):
    interact = make_match_interaction(matches, metadata, *args, **kwargs)
    interact.show_page()
    return interact


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.inspect_matches
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
# if False:
#     from __future__ import absolute_import, division, print_function
#     from six.moves import range
#     from plottool_ibeis import draw_func2 as df2
#     from plottool_ibeis import mpl_keypoint
#     from plottool_ibeis import mpl_sift  # NOQA
#     import vtool_ibeis.keypoint as ktool  # NOQA
#     import numpy as np
#     import matplotlib as mpl  # NOQA
#     from itertools import product as iprod
#     import vtool_ibeis  # NOQA
#     import ubelt as ub

#     TAU = 2 * np.pi  # References: tauday.com

#     # Hack these directions to be relative to gravity
#     RIGHT = ((0 * TAU / 4) - ktool.GRAVITY_THETA) % TAU
#     DOWN  = ((1 * TAU / 4) - ktool.GRAVITY_THETA) % TAU
#     LEFT  = ((2 * TAU / 4) - ktool.GRAVITY_THETA) % TAU
#     UP    = ((3 * TAU / 4) - ktool.GRAVITY_THETA) % TAU


#     def test_keypoint(xscale=1, yscale=1, ori=DOWN, skew=0):
#         # Test Keypoint
#         x, y = 0, 0
#         iv11, iv21, iv22 = xscale, skew, yscale
#         kp = np.array([x, y, iv11, iv21, iv22, ori])

#         # Test SIFT descriptor
#         sift = np.zeros(128)
#         sift[ 0: 8]   = 1
#         sift[ 8:16]   = .5
#         sift[16:24]   = .0
#         sift[24:32]   = .5
#         sift[32:40]   = .8
#         sift[40:48]   = .8
#         sift[48:56]   = .1
#         sift[56:64]   = .2
#         sift[64:72]   = .3
#         sift[72:80]   = .4
#         sift[80:88]   = .5
#         sift[88:96]   = .6
#         sift[96:104]  = .7
#         sift[104:112] = .8
#         sift[112:120] = .9
#         sift[120:128] = 1
#         sift = sift / np.sqrt((sift ** 2).sum())
#         sift = np.round(sift * 255)

#         kpts = np.array([kp])
#         sifts = np.array([sift])
#         return kpts, sifts


#     def square_axis(ax, s=3):
#         ax.set_xlim(-s, s)
#         ax.set_ylim(-s, s)
#         ax.set_aspect('equal')
#         ax.invert_yaxis()
#         df2.set_xticks([])
#         df2.set_yticks([])


#     def test_shape(ori=0, skew=0, xscale=1, yscale=1, pnum=(1, 1, 1), fnum=1):
#         df2.figure(fnum=fnum, pnum=pnum)
#         kpts, sifts = test_keypoint(ori=ori, skew=skew, xscale=xscale, yscale=yscale)
#         ax = df2.gca()
#         square_axis(ax)
#         mpl_keypoint.draw_keypoints(ax, kpts, sifts=sifts, ell_color=df2.ORANGE, ori=True,
#                                     rect_color=df2.DARK_RED,
#                                     ori_color=df2.DEEP_PINK, eig_color=df2.PINK,
#                                     rect=True, eig=True, bin_color=df2.RED,
#                                     arm1_color=df2.YELLOW, arm2_color=df2.BLACK)

#         kptsstr = '\n'.join(ktool.get_kpts_strs(kpts))
#         #print(kptsstr)
#         df2.upperleft_text(kptsstr)

#         title = 'xyscale=(%.1f, %.1f),\n skew=%.1f, ori=%.2ftau' % (xscale, yscale, skew, ori / TAU)
#         df2.set_title(title)
#         df2.dark_background()
#         return kpts, sifts


#     np.set_printoptions(precision=3)
#     px_ = 0

#     THETA1 = DOWN
#     THETA2 = (DOWN + DOWN + RIGHT) / 3
#     THETA3 = (DOWN + RIGHT) / 2
#     THETA4 = (DOWN + RIGHT + RIGHT) / 3
#     THETA5 = RIGHT

#     nRows = 2
#     nCols = 4


#     def pnum_(px=None):
#         global px_
#         if px is None:
#             px_ += 1
#             px = px_
#         return (nRows, nCols, px)

#     MIN_ORI = float(utool.argval('--min-ori', default=DOWN))
#     MAX_ORI = float(utool.argval('--max-ori', default=DOWN + TAU - .2))

#     MIN_X = .5
#     MAX_X = 2

#     MIN_SWEW = float(ub.argval('--min-skew', default=0))
#     MAX_SKEW = float(ub.argval('--max-skew', default=1))

#     MIN_Y = .5
#     MAX_Y = 2

#     kp_list = []

#     if __name__ == '__main__':

#         for row, col in iprod(range(nRows), range(nCols)):
#             #print((row, col))
#             alpha = col / (nCols)
#             beta  = row / (nRows)
#             xsca = (MIN_X    * (1 - alpha)) + (MAX_X    * (alpha))
#             ori  = (MIN_ORI  * (1 - alpha)) + (MAX_ORI  * (alpha))
#             skew = (MIN_SWEW * (1 - beta))  + (MAX_SKEW * (beta))
#             ysca = (MIN_Y    * (1 - beta))  + (MAX_Y    * (beta))

#             kpts, sifts = test_shape(pnum=pnum_(),
#                                      ori=ori,
#                                      skew=skew,
#                                      xscale=xsca,
#                                      yscale=ysca)
#         #print('+----')
#         #kp_list.append(kpts[0])
#         #S_list = ktool.get_ xy_axis_extents(kpts)
#         #print('xscale=%r yscale=%r, skew=%r' % (xsca, ysca, skew))
#         #print(S_list)

#         #scale_factor = 1
#         #offset = (0, 0)
#         #(_xs, _ys, _iv11s, _iv12s, _iv21s, _iv22s, _oris) = ktool.scaled_kpts(kpts, scale_factor, offset)

#         #invVR_aff2Ds = mpl_keypoint.get_invV_aff2Ds(_xs, _ys, _iv11s, _iv12s, _iv21s, _iv22s)
#         #aff = invVR_aff2Ds[0]

#         #ori = _oris[0]
#         #aff2 = mpl.transforms.Affine2D().rotate(-ori)

#         #print((aff + aff2).frozen())
#         #print((aff2 + aff).frozen())

#         #mpl_sift.draw_sift(ax, sift)
#         #df2.update()

#         exec(df2.present())



# -*- coding: utf-8 -*-
"""
Spatial verification of keypoint matches

Notation::
    1_m = img1_matches; 2_m = img2_matches
    x and y are locations, invV is the elliptical shapes.
    fx are the original feature indexes (used for making sure 1 keypoint isn't assigned to 2)

Look Into::
    Standard
    skimage.transform
    http://stackoverflow.com/questions/11462781/fast-2d-rigid-body-transformations-in-numpy-scipy
    skimage.transform.fast_homography(im, H)

FIXME:
    is it scaled_thresh or scaled_thresh_sqrd

References:
    http://ags.cs.uni-kl.de/fileadmin/inf_ags/3dcv-ws11-12/3DCV_WS11-12_lec04.pdf
    http://www.imgfsr.com/CVPR2011/Tutorial6/RANSAC_CVPR2011.pdf
    http://szeliski.org/Book/drafts/SzeliskiBook_20100903_draft.pdf Page 317

Notes:
    Invariants of affine transforms - parallel lines, ratios of parallel lengths, ratios of areas
    Invariants of homographies - cross‐ratio of four points on a line (ratio of ratio)
"""
from __future__ import absolute_import, division, print_function
from six.moves import range
import warnings  # NOQA
import six  # NOQA
import utool as ut
import numpy as np
import numpy.linalg as npl
import scipy.sparse as sps
import scipy.sparse.linalg as spsl
import operator as op
# from numpy.core.umath_tests import matrix_multiply
import vtool_ibeis.keypoint as ktool
import vtool_ibeis.linalg as ltool
import vtool_ibeis.distance
import cv2
from .util_math import TAU

try:
    from vtool_ibeis import sver_c_wrapper
    HAVE_SVER_C_WRAPPER = not ut.get_argflag('--no-c')
except Exception as ex:
    HAVE_SVER_C_WRAPPER = False
    if ut.VERBOSE:
        ut.printex(ex, 'please build the sver c wrapper (run with --rebuild-sver')
    if False:
        raise


VERBOSE_SVER = ut.get_argflag('--verb-sver')

SV_DTYPE = np.float64
INDEX_DTYPE = np.int32


def build_lstsqrs_Mx9(xy1_mn, xy2_mn):
    """ Builds the M x 9 least squares matrix

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-build_lstsqrs_Mx9

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> kpts1, kpts2 = demodata.get_dummy_kpts_pair()
        >>> xy1_mn = ktool.get_xys(kpts1).astype(np.float64)
        >>> xy2_mn = ktool.get_xys(kpts2).astype(np.float64)
        >>> Mx9 = build_lstsqrs_Mx9(xy1_mn, xy2_mn)
        >>> import ubelt as ub
        >>> result = (ub.repr2(Mx9[0:2], suppress_small=True, precision=2, with_dtype=True))
        >>> print(result)
        np.array([[  0.00e+00,   0.00e+00,   0.00e+00,  -3.20e+01,  -2.72e+01,
                    -1.00e+00,   8.82e+02,   7.49e+02,   2.76e+01],
                  [  3.20e+01,   2.72e+01,   1.00e+00,   0.00e+00,   0.00e+00,
                     0.00e+00,  -1.09e+03,  -9.28e+02,  -3.42e+01]], dtype=np.float64)

    References:
        http://dip.sun.ac.za/~stefan/TW793/attach/notes/homography_estimation.pdf
        http://szeliski.org/Book/drafts/SzeliskiBook_20100903_draft.pdf Page 317
        http://vision.ece.ucsb.edu/~zuliani/Research/RANSAC/docs/RANSAC4Dummies.pdf page 53
    """
    x1_mn = xy1_mn[0]
    y1_mn = xy1_mn[1]
    x2_mn = xy2_mn[0]
    y2_mn = xy2_mn[1]
    num_pts = x1_mn.shape[0]
    #Mx9 = np.zeros((2 * num_pts, 9), dtype=SV_DTYPE)
    Mx9 = np.empty((2 * num_pts, 9), dtype=SV_DTYPE)
    for ix in range(num_pts):  # Loop over inliers
        # Concatenate all 2x9 matrices into an Mx9 matrix
        u2        = x2_mn[ix]
        v2        = y2_mn[ix]
        x1        = x1_mn[ix]
        y1        = y1_mn[ix]
        (d, e, f) = (     -x1,      -y1,  -1)
        (g, h, i) = ( v2 * x1,  v2 * y1,  v2)
        (j, k, l) = (      x1,       y1,   1)
        (p, q, r) = (-u2 * x1, -u2 * y1, -u2)
        Mx9[ix * 2]     = (0, 0, 0, d, e, f, g, h, i)
        Mx9[ix * 2 + 1] = (j, k, l, 0, 0, 0, p, q, r)
    return Mx9


def try_svd(M):
    """
    CommandLine:
        python -m vtool_ibeis.spatial_verification try_svd

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> rng = np.random.RandomState(42)
        >>> num = 1000
        >>> xy1_mn = rng.randn(2, num)
        >>> xy2_mn = rng.randn(2, num)
        >>> M = build_lstsqrs_Mx9(xy1_mn, xy2_mn)
        >>> print('M.shape = %r' % (M.shape,))
        >>> USV = npl.svd(M, full_matrices=True, compute_uv=True)
        >>> USV = try_svd(M)

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> num = np.ceil(np.sqrt(2000))
        >>> kpts1, kpts2 = demodata.get_dummy_kpts_pair(wh_num=(num, num))
        >>> xy1_mn = ktool.get_xys(kpts1).astype(np.float64)
        >>> xy2_mn = ktool.get_xys(kpts2).astype(np.float64)
        >>> M = build_lstsqrs_Mx9(xy1_mn, xy2_mn)
        >>> print('M.shape = %r' % (M.shape,))
        >>> USV = npl.svd(M, full_matrices=True, compute_uv=True)
        >>> USV = try_svd(M)
    """
    #if M.shape[0] > 2500:
    #    # hack to prevent bug in lapack
    #    M = M[:2500]
    try:
        USV = npl.svd(M, full_matrices=True, compute_uv=True)
    except MemoryError as ex:
        ut.printex(ex, '[sver] Caught MemErr during full SVD. Trying sparse SVD.')
        M_sparse = sps.lil_matrix(M)
        USV = spsl.svds(M_sparse)
    except npl.LinAlgError as ex:
        ut.printex(ex, '[sver] svd did not converge')
        raise
    except Exception as ex:
        ut.printex(ex, '[sver] svd error')
        raise
    return USV


def build_affine_lstsqrs_Mx6(xy1_man, xy2_man):
    """
    CURRENTLY NOT WORKING

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-build_affine_lstsqrs_Mx6

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> kpts1, kpts2 = demodata.get_dummy_kpts_pair()
        >>> xy1_man = ktool.get_xys(kpts1).astype(np.float64)
        >>> xy2_man = ktool.get_xys(kpts2).astype(np.float64)
        >>> Mx6 = build_affine_lstsqrs_Mx6(xy1_man, xy2_man)
        >>> import ubelt as ub
        >>> print(ub.repr2(Mx6))
        >>> result = ut.hashstr(Mx6)
        >>> print(result)

    Sympy:
        import sympy as sym
        x1, y1, x2, y2 = sym.symbols('x1, y1, x2, y2')
        A = sym.Matrix([
            [x1, y1,  0,  0, 1, 0],
            [ 0,  0, x1, y1, 0, 1],
        ])
        b = sym.Matrix([[x2], [y2]])
        x = (A.T.multiply(A)).inv().multiply(A.T.multiply(b))
        x = (A.T.multiply(A)).pinv().multiply(A.T.multiply(b))

    References:
        https://www.cs.ubc.ca/~lowe/papers/ijcv04.pdf page 22
    """
    x1_mn = xy1_man[0]
    y1_mn = xy1_man[1]
    x2_mn = xy2_man[0]
    y2_mn = xy2_man[1]
    num_pts = x1_mn.shape[0]
    Mx6 = np.empty((2 * num_pts, 6), dtype=SV_DTYPE)
    b = np.empty((2 * num_pts, 1), dtype=SV_DTYPE)
    for ix in range(num_pts):  # Loop over inliers
        # Concatenate all 2x9 matrices into an Mx6 matrix
        x1 = x1_mn[ix]
        x2 = x2_mn[ix]
        y1 = y1_mn[ix]
        y2 = y2_mn[ix]
        Mx6[ix * 2]     = (x1, y1,  0,  0,  1,  0)
        Mx6[ix * 2 + 1] = ( 0,  0, x1, y1,  0,  1)
        b[ix * 2] = x2
        b[ix * 2 + 1] = y2

    #npl.solve(b, Mx6)
    U, s, Vt = try_svd(Mx6)

    if False:
        S_ = np.zeros((len(U), len(Vt)))
        S_[np.diag_indices(len(s))] = s
        U.dot(S_).dot(Vt)
        assert np.allclose(U.dot(S_).dot(Vt), Mx6)
        assert not np.allclose(U.dot(S_).dot(Vt.T), Mx6)

    # Inefficient, but I think the math works
    # We want to solve Ax=b (where A is the Mx6 in this case)
    # Ax = b
    # (U S V.T) x = b
    # x = (U.T inv(S) V) b
    Sinv = np.zeros((len(Vt), len(U)))
    Sinv[np.diag_indices(len(s))] = 1 / s
    a = Vt.T.dot(Sinv).dot(U.T).dot(b).T[0]
    A = np.array([
        [a[0], a[1], a[4]],
        [a[2], a[3], a[5]],
        [   0,    0,    1],
    ])
    return A
    # TODO FIXME
    #return Mx6


def compute_affine(xy1_man, xy2_man):
    """
    Args:
        xy1_mn (ndarray[ndim=2]): xy points in image1
        xy2_mn (ndarray[ndim=2]): corresponding xy points in image 2

    Returns:
        ndarray[shape=(3,3)]: A - affine matrix

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-compute_affine:1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> import vtool_ibeis.keypoint as ktool
        >>> kpts1, kpts2 = demodata.get_dummy_kpts_pair()
        >>> xy1_mn = ktool.get_xys(kpts1)
        >>> xy2_mn = ktool.get_xys(kpts2)
        >>> A = compute_affine(xy1_mn, xy1_mn)
        >>> result =str(A)
        >>> result = np.array_str(A, precision=2)
        >>> print(result)

    Example1:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> import vtool_ibeis.keypoint as ktool
        >>> xy1_man, xy2_man, rchip1, rchip2, T1, T2 = testdata_matching_affine_inliers_normalized()
        >>> A_prime = compute_affine(xy1_man, xy2_man)
        >>> A = npl.solve(T2, A_prime).dot(T1)
        >>> A /= A[2, 2]
        >>> result = np.array_str(A, precision=2)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> rchip2_blendA = pt.draw_sv.get_blended_chip(rchip1, rchip2, A)
        >>> pt.imshow(rchip2_blendA)
        >>> ut.show_if_requested()
        [[  1.19e+00  -1.06e-02  -4.49e+01]
         [ -2.22e-01   1.12e+00  -2.78e+01]
         [  0.00e+00   0.00e+00   1.00e+00]]
    """
    # Solve for the nullspace of the Mx6 matrix (solves least squares)
    #Mx6 = build_affine_lstsqrs_Mx6(xy1_man, xy2_man)
    A = build_affine_lstsqrs_Mx6(xy1_man, xy2_man)
    #U, S, V = try_svd(Mx6)
    #a = V[5]  # Hack for Cython.wraparound(False)
    #a = a / a[-1]
    #A = np.array([
    #    [a[0], a[1], a[4]],
    #    [a[2], a[3], a[5]],
    #    [   0,    0,    1],
    #])
    return A


def compute_homog(xy1_mn, xy2_mn):
    """
    Generate 6 degrees of freedom homography transformation
    Computes homography from normalized (0 to 1) point correspondences
    from 2 --> 1
    (database->query)

    Args:
        xy1_mn (ndarray[ndim=2]): xy points in image1
        xy2_mn (ndarray[ndim=2]): corresponding xy points in image 2

    Returns:
        ndarray[shape=(3,3)]: H - homography matrix

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-compute_homog:1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.keypoint as ktool
        >>> import vtool_ibeis.demodata as demodata
        >>> kpts1, kpts2 = demodata.get_dummy_kpts_pair()
        >>> xy1_mn = ktool.get_xys(kpts1)
        >>> xy2_mn = ktool.get_xys(kpts2)
        >>> H = compute_homog(xy1_mn, xy2_mn)
        >>> #result = ut.hashstr(H)
        >>> result = np.array_str(H, precision=2)
        >>> print(result)
        [[  1.83e-03   2.85e-03  -7.11e-01]
         [  2.82e-03   1.80e-03  -7.03e-01]
         [  1.67e-05   1.68e-05  -5.53e-03]]

    Example1:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.keypoint as ktool
        >>> xy1_man, xy2_man, rchip1, rchip2, T1, T2 = testdata_matching_affine_inliers_normalized()
        >>> H_prime = compute_homog(xy1_man, xy2_man)
        >>> H = npl.solve(T2, H_prime).dot(T1)
        >>> H = H / H[2, 2]
        >>> result = np.array_str(H, precision=2)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> rchip2_blendH = pt.draw_sv.get_blended_chip(rchip1, rchip2, H)
        >>> pt.imshow(rchip2_blendH)
        >>> ut.show_if_requested()
        [[  9.22e-01  -2.50e-01   2.75e+01]
         [ -2.04e-01   8.79e-01  -7.94e+00]
         [ -1.82e-04  -5.99e-04   1.00e+00]]
    """
    # Solve for the nullspace of the Mx9 matrix (solves least squares)
    Mx9 = build_lstsqrs_Mx9(xy1_mn, xy2_mn)
    U, S, V = try_svd(Mx9)
    # Rearange the nullspace into a homography
    #h = V[-1]  # v = V.H
    h = V[8]  # Hack for Cython.wraparound(False)
    # FIXME: THERE IS A BUG HERE, sometimes len(V) = 6. why???
    H = np.vstack((h[0:3], h[3:6], h[6:9]))
    return H


def testdata_matching_affine_inliers():
    import vtool_ibeis.demodata as demodata
    import vtool_ibeis as vt
    scale_thresh = 2.0
    xy_thresh = ut.get_argval('--xy-thresh', type_=float, default=.01)
    dlen_sqrd2 = 447271.015
    ori_thresh = 1.57
    xy_thresh_sqrd = dlen_sqrd2 * xy_thresh
    featkw = ut.argparse_dict(vt.get_extract_features_default_params())
    fname1 = ut.get_argval('--fname1', type_=str, default='tsukuba_l')
    fname2 = ut.get_argval('--fname2', type_=str, default='tsukuba_r')
    (kpts1, kpts2, fm, fs, rchip1, rchip2) = demodata.testdata_ratio_matches(fname1, fname2, **featkw)
    aff_inliers, aff_errors, Aff = get_best_affine_inliers_(
        kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh, ori_thresh)
    return kpts1, kpts2, fm, aff_inliers, rchip1, rchip2, xy_thresh_sqrd


def testdata_matching_affine_inliers_normalized():
    tup = testdata_matching_affine_inliers()
    kpts1, kpts2, fm, aff_inliers, rchip1, rchip2, xy_thresh_sqrd = tup
    kpts1_ma = kpts1.take(fm.T[0].take(aff_inliers), axis=0)
    kpts2_ma = kpts2.take(fm.T[1].take(aff_inliers), axis=0)
    #kpts1_ma, kpts2_ma, rchip1, rchip2, xy_thresh_sqrd = testdata_matching_affine_inliers()
    # Matching affine inliers
    xy1_ma = ktool.get_xys(kpts1_ma)
    xy2_ma = ktool.get_xys(kpts2_ma)
    # Matching affine inliers normalized
    xy1_man, T1 = ltool.whiten_xy_points(xy1_ma)
    xy2_man, T2 = ltool.whiten_xy_points(xy2_ma)
    return xy1_man, xy2_man, rchip1, rchip2, T1, T2


def _test_hypothesis_inliers(Aff, invVR1s_m, xy2_m, det2_m, ori2_m,
                             xy_thresh_sqrd, scale_thresh_sqrd, ori_thresh):
    """
    Critical section code. Inner loop of _test_hypothesis_inliers

    Returns:
        tuple: hypo_inliers, hypo_errors

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-_test_hypothesis_inliers

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> from vtool_ibeis.spatial_verification import _test_hypothesis_inliers  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> import vtool_ibeis.keypoint as ktool
        >>> _kw1 = dict(seed=12, damping=1.2, wh_stride=(30, 30))
        >>> _kw2 = dict(seed=24, damping=1.6, wh_stride=(30, 30))
        >>> kpts1 = demodata.perterbed_grid_kpts(**_kw1).astype(np.float64)
        >>> kpts2 = demodata.perterbed_grid_kpts(**_kw2).astype(np.float64)
        >>> fm = demodata.make_dummy_fm(len(kpts1)).astype(np.int32)
        >>> kpts1_m = kpts1[fm.T[0]]
        >>> kpts2_m = kpts2[fm.T[1]]
        >>> xy_thresh_sqrd = np.float64(.009) ** 2
        >>> scale_thresh_sqrd = np.float64(2)
        >>> ori_thresh = np.float64(TAU / 4)
        >>> # Get keypoints to project in matrix form
        >>> #invVR1s_m = ktool.get_invV_mats(kpts1_m, with_trans=True, with_ori=True)
        >>> #print(invVR1s_m[0])
        >>> invVR1s_m = ktool.get_invVR_mats3x3(kpts1_m)
        >>> RV1s_m = ktool.get_RV_mats_3x3(kpts1_m)
        >>> invVR2s_m = ktool.get_invVR_mats3x3(kpts2_m)
        >>> # The transform from kp1 to kp2 is given as:
        >>> Aff_mats = op.matmul(invVR2s_m, RV1s_m)
        >>> Aff = Aff_mats[0]
        >>> # Get components to test projects against
        >>> xy2_m  = ktool.get_invVR_mats_xys(invVR2s_m)
        >>> det2_m = ktool.get_sqrd_scales(kpts2_m)
        >>> ori2_m = ktool.get_invVR_mats_oris(invVR2s_m)
        >>> output = _test_hypothesis_inliers(Aff, invVR1s_m, xy2_m, det2_m, ori2_m, xy_thresh_sqrd, scale_thresh_sqrd, ori_thresh)
        >>> import ubelt as ub
        >>> output_str = ub.repr2(output, precision=2, suppress_small=True)
        >>> print('output_str = %s' % (output_str,))
        >>> hypo_inliers, hypo_errors = output
        >>> # Inverting matrices is different in python2
        >>> result = 'nInliers=%r hash=%s' % (len(hypo_inliers), ut.hash_data(output))
        >>> print(result)
    """
    # Map keypoints from image 1 onto image 2
    invVR1s_mt = op.matmul(Aff, invVR1s_m)

    # Get projection components
    _xy1_mt   = ktool.get_invVR_mats_xys(invVR1s_mt)
    _det1_mt  = ktool.get_invVR_mats_sqrd_scale(invVR1s_mt)
    _ori1_mt  = ktool.get_invVR_mats_oris(invVR1s_mt)
    ## Check for projection errors
    xy_err    = vtool_ibeis.distance.L2_sqrd(xy2_m.T, _xy1_mt.T, dtype=SV_DTYPE)
    scale_err = vtool_ibeis.distance.det_distance(_det1_mt, det2_m)
    ori_err   = vtool_ibeis.distance.ori_distance(_ori1_mt, ori2_m)

    # Mark keypoints which are inliers to this hypothosis
    xy_inliers_flag    = np.less(xy_err, xy_thresh_sqrd)
    scale_inliers_flag = np.less(scale_err, scale_thresh_sqrd)
    ori_inliers_flag   = np.less(ori_err, ori_thresh)
    #np.logical_and(xy_inliers_flag, scale_inliers_flag)
    # TODO Add uniqueness of matches constraint
    #hypo_inliers_flag = np.empty(xy_inliers_flag.size, dtype=bool)
    hypo_inliers_flag = xy_inliers_flag  # Try to re-use memory
    np.logical_and(hypo_inliers_flag, ori_inliers_flag, out=hypo_inliers_flag)
    np.logical_and(hypo_inliers_flag, scale_inliers_flag, out=hypo_inliers_flag)
    #hypo_inliers_flag = np.logical_and.reduce(
    #    [xy_inliers_flag, ori_inliers_flag, scale_inliers_flag])
    # this is also slower
    hypo_inliers = np.where(hypo_inliers_flag)[0]
    hypo_errors = (xy_err, ori_err, scale_err)
    return hypo_inliers, hypo_errors


def get_affine_inliers(kpts1, kpts2, fm, fs,
                        xy_thresh_sqrd,
                        scale_thresh_sqrd,
                        ori_thresh):
    """
    Estimates inliers deterministically using elliptical shapes

    Compute all transforms from kpts1 to kpts2 (enumerate all hypothesis)
    We transform from chip1 -> chip2
    The determinants are squared keypoint scales

    Returns:
        tuple: aff_inliers_list, aff_errors_list, Aff_mats

    Notes:
        FROM PERDOCH 2009::
            H = inv(Aj).dot(Rj.T).dot(Ri).dot(Ai)
            H = inv(Aj).dot(Ai)
            The input invVs = perdoch.invA's

    CommandLine:
        python2 -m vtool_ibeis.spatial_verification --test-get_affine_inliers
        python3 -m vtool_ibeis.spatial_verification --test-get_affine_inliers

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> import vtool_ibeis.keypoint as ktool
        >>> kpts1, kpts2 = demodata.get_dummy_kpts_pair((100, 100))
        >>> fm = demodata.make_dummy_fm(len(kpts1)).astype(np.int32)
        >>> fs = np.ones(len(fm), dtype=np.float64)
        >>> xy_thresh_sqrd = ktool.KPTS_DTYPE(.009) ** 2
        >>> scale_thresh_sqrd = ktool.KPTS_DTYPE(2)
        >>> ori_thresh = ktool.KPTS_DTYPE(TAU / 4)
        >>> output = get_affine_inliers(kpts1, kpts2, fm, fs, xy_thresh_sqrd,
        >>>                             scale_thresh_sqrd, ori_thresh)
        >>> output_str = ut.repr3(output, precision=2, suppress_small=True)
        >>> print('output_str = %s' % (output_str,))
        >>> aff_inliers_list, aff_errors_list, Aff_mats = output
        >>> result = 'nInliers=%r hash=%s' % (len(aff_inliers_list), ut.hash_data(output_str))
        >>> print(result)
    """
    #http://ipython-books.github.io/featured-01/
    kpts1_m = kpts1.take(fm.T[0], axis=0)
    kpts2_m = kpts2.take(fm.T[1], axis=0)

    # Get keypoints to project in matrix form
    #invVR2s_m = ktool.get_invV_mats(kpts2_m, with_trans=True, with_ori=True)
    #invVR1s_m = ktool.get_invV_mats(kpts1_m, with_trans=True, with_ori=True)
    invVR2s_m = ktool.get_invVR_mats3x3(kpts2_m)
    invVR1s_m = ktool.get_invVR_mats3x3(kpts1_m)
    RV1s_m    = ktool.invert_invV_mats(invVR1s_m)  # 539 us
    # BUILD ALL HYPOTHESIS TRANSFORMS: The transform from kp1 to kp2 is:
    Aff_mats = op.matmul(invVR2s_m, RV1s_m)
    # Get components to test projects against
    xy2_m  = ktool.get_xys(kpts2_m)
    det2_m = ktool.get_sqrd_scales(kpts2_m)
    ori2_m = ktool.get_oris(kpts2_m)
    # SLOWER EQUIVALENT
    # RV1s_m    = ktool.get_V_mats(kpts1_m, with_trans=True, with_ori=True)  # 5.2 ms
    # xy2_m  = ktool.get_invVR_mats_xys(invVR2s_m)
    # ori2_m = ktool.get_invVR_mats_oris(invVR2s_m)
    # assert np.all(ktool.get_oris(kpts2_m) == ktool.get_invVR_mats_oris(invVR2s_m))
    # assert np.all(ktool.get_xys(kpts2_m) == ktool.get_invVR_mats_xys(invVR2s_m))

    # The previous versions of this function were all roughly comparable.
    # The for loop one was the slowest. I'm deciding to go with the one
    # where there is no internal function definition. It was moderately faster,
    # and it gives us access to profile that function
    inliers_and_errors_list = [_test_hypothesis_inliers(Aff, invVR1s_m, xy2_m,
                                                        det2_m, ori2_m,
                                                        xy_thresh_sqrd,
                                                        scale_thresh_sqrd,
                                                        ori_thresh)
                               for Aff in Aff_mats]
    aff_inliers_list = [tup[0] for tup in inliers_and_errors_list]
    aff_errors_list  = [tup[1] for tup in inliers_and_errors_list]
    return aff_inliers_list, aff_errors_list, Aff_mats


def get_best_affine_inliers(kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh,
                            ori_thresh, forcepy=False):
    """ Tests each hypothesis and returns only the best transformation and inliers
    """
    # Test each affine hypothesis
    # get list if inliers, errors, the affine matrix for each hypothesis
    if HAVE_SVER_C_WRAPPER and not forcepy:
        aff_inliers_list, aff_errors_list, Aff_mats = sver_c_wrapper.get_affine_inliers_cpp(
            kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh, ori_thresh)
    else:
        aff_inliers_list, aff_errors_list, Aff_mats = get_affine_inliers(kpts1, kpts2,
                                                                         fm, fs,
                                                                         xy_thresh_sqrd,
                                                                         scale_thresh,
                                                                         ori_thresh)

    #ut.embed()
    # Determine the best hypothesis using the number of inliers
    # TODO: other measures in the error lists could be used as well
    #nInliers_list = np.array([len(inliers) for inliers in aff_inliers_list])

    weight_list = np.array([fs.take(inliers).sum() for inliers in aff_inliers_list])
    #ut.embed()
    #sortx = weight_list.argsort()[::-1]  # sort by non-inliers
    #best_index = sortx[0]  # chose best
    best_index = weight_list.argmax()
    aff_inliers = aff_inliers_list[best_index]
    aff_errors = aff_errors_list[best_index]
    Aff = Aff_mats[best_index]
    return aff_inliers, aff_errors, Aff


def get_normalized_affine_inliers(kpts1, kpts2, fm, aff_inliers):
    """
    returns xy-inliers that are normalized to have a mean of 0 and std of 1 as
    well as the transformations so the inverse can be taken
    """
    fm_affine = fm.take(aff_inliers, axis=0)
    # Get corresponding points and shapes
    kpts1_ma = kpts1.take(fm_affine.T[0], axis=0)
    kpts2_ma = kpts2.take(fm_affine.T[1], axis=0)
    #kpts1_ma = kpts1.take(fm_affine.T[0], axis=0)
    #kpts2_ma = kpts2.take(fm_affine.T[1], axis=0)
    # Normalize affine inliers xy locations
    xy1_ma = ktool.get_xys(kpts1_ma)
    xy2_ma = ktool.get_xys(kpts2_ma)
    xy1_man, T1 = ltool.whiten_xy_points(xy1_ma)
    xy2_man, T2 = ltool.whiten_xy_points(xy2_ma)
    return xy1_man, xy2_man, T1, T2


def unnormalize_transform(M_prime, T1, T2):
    # Then compute ax = b  [aka: x = npl.solve(a, b)]
    M = npl.solve(T2, M_prime).dot(T1)  # Unnormalize
    # homographies that only differ by a scale factor are equivalent
    M /= M[2, 2]
    return M


def estimate_refined_transform(kpts1, kpts2, fm, aff_inliers, refine_method='homog'):
    """ estimates final transformation using normalized affine inliers

    References:
        http://docs.opencv.org/2.4/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html
    """
    xy1_man, xy2_man, T1, T2 = get_normalized_affine_inliers(kpts1, kpts2, fm, aff_inliers)
    # Compute homgraphy transform from chip1 -> chip2 using affine inliers
    # homographys assume the two images are planar, or the camera is
    # rotating around the subject

    if refine_method == 'homog':
        H_prime = compute_homog(xy1_man, xy2_man)
    elif refine_method == 'affine':
        H_prime = compute_affine(xy1_man, xy2_man)
    elif refine_method == 'cv2-homog':
        H_prime, mask = cv2.findHomography(xy1_man.T, xy2_man.T, method=0)
    elif refine_method == 'cv2-ransac-homog':
        H_prime, mask = cv2.findHomography(
            xy1_man.T, xy2_man.T, method=cv2.RANSAC,
            ransacReprojThreshold=3)
    elif refine_method == 'cv2-lmeds-homog':
        H_prime, mask = cv2.findHomography(xy1_man.T, xy2_man.T, method=cv2.LMEDS)
    #elif refine_method == 'fund':
    #    # the fundamental matrix only implies: x'.T.dot(F).dot(x) == 0
    #    # it maps a point from one image onto a line in the second image.
    #    H_prime = cv2.findFundamentalMat(xy1_man.T, xy2_man.T, method=cv2.FM_LMEDS)[0]
    #    H_prime = cv2.findFundamentalMat(xy1_man.T, xy2_man.T, method=cv2.FM_8POINT)[0]
    else:
        raise NotImplementedError('[vtool_ibeis] Unknown refine_method=%r' % (refine_method,))

    #H_prime /= H_prime[2, 2]
    # Different methods?
    #H_prime = compute_affine(xy1_man, xy2_man)
    #import cv2
    #H_prime = cv2.findHomography(xy1_man.T, xy2_man.T)[0]
    #H = compute_affine(xy1_ma, xy2_ma)
    #print(H)
    H = unnormalize_transform(H_prime, T1, T2)
    rank = npl.matrix_rank(H)
    #print(rank)
    if rank != 3:
        raise npl.LinAlgError('Rank defficient homography ')
    return H


def test_homog_errors(H, kpts1, kpts2, fm, xy_thresh_sqrd, scale_thresh,
                      ori_thresh, full_homog_checks=True):
    r"""
    Test to see which keypoints the homography correctly maps

    Args:
        H (ndarray[float64_t, ndim=2]):  homography/perspective matrix
        kpts1 (ndarray[float32_t, ndim=2]):  keypoints
        kpts2 (ndarray[float32_t, ndim=2]):  keypoints
        fm (list):  list of feature matches as tuples (qfx, dfx)
        xy_thresh_sqrd (float):
        scale_thresh (float):
        ori_thresh (float):  angle in radians
        full_homog_checks (bool):

    Returns:
        tuple: homog_tup1

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:0 --show
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:0 --show --rotation_invariance
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:0 --show --rotation_invariance --no-affine-invariance --xy-thresh=.001
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:0 --show --rotation_invariance --no-affine-invariance --xy-thresh=.001 --no-full-homog-checks
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:0 --show --no-full-homog-checks
        # --------------
        # Shows (sorta) how inliers are computed
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:1 --show
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:1 --show --rotation_invariance
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:1 --show --rotation_invariance --no-affine-invariance --xy-thresh=.001
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:1 --show --rotation_invariance --xy-thresh=.001
        python -m vtool_ibeis.spatial_verification --test-test_homog_errors:0 --show --rotation_invariance --xy-thresh=.001

    Example0:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> kpts1, kpts2, fm, aff_inliers, rchip1, rchip2, xy_thresh_sqrd = testdata_matching_affine_inliers()
        >>> H = estimate_refined_transform(kpts1, kpts2, fm, aff_inliers)
        >>> scale_thresh, ori_thresh = 2.0, 1.57
        >>> full_homog_checks = not ut.get_argflag('--no-full-homog-checks')
        >>> homog_tup1 = test_homog_errors(H, kpts1, kpts2, fm, xy_thresh_sqrd, scale_thresh, ori_thresh, full_homog_checks)
        >>> homog_tup = (homog_tup1[0], homog_tup1[2])
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.draw_sv.show_sv(rchip1, rchip2, kpts1, kpts2, fm, homog_tup=homog_tup)
        >>> ut.show_if_requested()

    Example1:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> kpts1, kpts2, fm_, aff_inliers, rchip1, rchip2, xy_thresh_sqrd = testdata_matching_affine_inliers()
        >>> H = estimate_refined_transform(kpts1, kpts2, fm_, aff_inliers)
        >>> scale_thresh, ori_thresh = 2.0, 1.57
        >>> full_homog_checks = not ut.get_argflag('--no-full-homog-checks')
        >>> # ----------------
        >>> # Take subset of feature matches
        >>> fm = fm_
        >>> scale_err, xy_err, ori_err = \
        ...     ut.exec_func_src(test_homog_errors, globals(), locals(),
        ...     'scale_err, xy_err, ori_err'.split(', '))
        >>> # we only care about checking out scale and orientation here. ignore bad xy points
        >>> xy_inliers_flag = np.less(xy_err, xy_thresh_sqrd)
        >>> scale_err[~xy_inliers_flag] = 0
        >>> # filter
        >>> fm = fm_[np.array(scale_err).argsort()[::-1][:10]]
        >>> fm = fm_[np.array(scale_err).argsort()[::-1][:10]]
        >>> # Exec sourcecode
        >>> kpts1_m, kpts2_m, off_xy1_m, off_xy1_mt, dxy1_m, dxy1_mt, xy2_m, xy1_m, xy1_mt, scale_err, xy_err, ori_err = \
        ...     ut.exec_func_src(test_homog_errors, globals(), locals(),
        ...     'kpts1_m, kpts2_m, off_xy1_m, off_xy1_mt, dxy1_m, dxy1_mt, xy2_m, xy1_m, xy1_mt, scale_err, xy_err, ori_err'.split(', '))
        >>> #---------------
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.figure(fnum=1, pnum=(1, 2, 1), title='orig points and offset point')
        >>> segments_list1 = np.array(list(zip(xy1_m.T.tolist(), off_xy1_m.T.tolist())))
        >>> pt.draw_line_segments(segments_list1, color=pt.LIGHT_BLUE)
        >>> pt.dark_background()
        >>> #---------------
        >>> pt.figure(fnum=1, pnum=(1, 2, 2), title='transformed points and matching points')
        >>> #---------------
        >>> # first have to make corresponding offset points
        >>> # Use reference point for scale and orientation tests
        >>> oris2_m   = ktool.get_oris(kpts2_m)
        >>> scales2_m = ktool.get_scales(kpts2_m)
        >>> dxy2_m    = np.vstack((np.sin(oris2_m), -np.cos(oris2_m)))
        >>> scaled_dxy2_m = dxy2_m * scales2_m[None, :]
        >>> off_xy2_m = xy2_m + scaled_dxy2_m
        >>> # Draw transformed semgents
        >>> segments_list2 = np.array(list(zip(xy2_m.T.tolist(), off_xy2_m.T.tolist())))
        >>> pt.draw_line_segments(segments_list2, color=pt.GREEN)
        >>> # Draw corresponding matches semgents
        >>> segments_list3 = np.array(list(zip(xy1_mt.T.tolist(), off_xy1_mt.T.tolist())))
        >>> pt.draw_line_segments(segments_list3, color=pt.RED)
        >>> # Draw matches between correspondences
        >>> segments_list4 = np.array(list(zip(xy1_mt.T.tolist(), xy2_m.T.tolist())))
        >>> pt.draw_line_segments(segments_list4, color=pt.ORANGE)
        >>> pt.dark_background()
        >>> #---------------
        >>> #vt.get _xy_axis_extents(kpts1_m)
        >>> #pt.draw_sv.show_sv(rchip1, rchip2, kpts1, kpts2, fm, homog_tup=homog_tup)
        >>> ut.show_if_requested()
    """
    kpts1_m = kpts1.take(fm.T[0], axis=0)
    kpts2_m = kpts2.take(fm.T[1], axis=0)
    # Transform all xy1 matches to xy2 space
    xy1_m   = ktool.get_xys(kpts1_m)
    #with ut.embed_on_exception_context:
    xy1_mt  = ltool.transform_points_with_homography(H, xy1_m)
    #xy1_mt  = ktool.transform_kpts_xys(H, kpts1_m)
    xy2_m   = ktool.get_xys(kpts2_m)
    # --- Find (Squared) Homography Distance Error ---
    # You cannot test for scale or orientation easily here because
    # you no longer have an ellipse? (maybe, probably have a conic) when using a
    # projective transformation
    xy_err = vtool_ibeis.distance.L2_sqrd(xy1_mt.T, xy2_m.T)
    # Estimate final inliers
    #ut.embed()
    if full_homog_checks:
        # TODO: may need to use more than one reference point
        # Use reference point for scale and orientation tests
        oris1_m   = ktool.get_oris(kpts1_m)
        scales1_m = ktool.get_scales(kpts1_m)
        # Get point offsets with unit length
        dxy1_m    = np.vstack((np.sin(oris1_m), -np.cos(oris1_m)))
        scaled_dxy1_m = dxy1_m * scales1_m[None, :]
        off_xy1_m = xy1_m + scaled_dxy1_m
        # transform reference point
        off_xy1_mt = ltool.transform_points_with_homography(H, off_xy1_m)
        scaled_dxy1_mt = xy1_mt - off_xy1_mt
        scales1_mt = npl.norm(scaled_dxy1_mt, axis=0)
        #with warnings.catch_warnings():
        #    warnings.simplefilter("ignore")
        dxy1_mt = scaled_dxy1_mt / scales1_mt
        # adjust for gravity vector being 0
        oris1_mt = np.arctan2(dxy1_mt[1], dxy1_mt[0]) - ktool.GRAVITY_THETA
        _det1_mt = scales1_mt ** 2
        det2_m = ktool.get_sqrd_scales(kpts2_m)
        ori2_m = ktool.get_oris(kpts2_m)
        #xy_err    = vtool_ibeis.distance.L2_sqrd(xy2_m.T, _xy1_mt.T)
        scale_err = vtool_ibeis.distance.det_distance(_det1_mt, det2_m)
        ori_err   = vtool_ibeis.distance.ori_distance(oris1_mt, ori2_m)
        ###
        xy_inliers_flag = np.less(xy_err, xy_thresh_sqrd)
        scale_inliers_flag = np.less(scale_err, scale_thresh)
        ori_inliers_flag   = np.less(ori_err, ori_thresh)
        hypo_inliers_flag = xy_inliers_flag  # Try to re-use memory
        np.logical_and(hypo_inliers_flag, ori_inliers_flag, out=hypo_inliers_flag)
        np.logical_and(hypo_inliers_flag, scale_inliers_flag, out=hypo_inliers_flag)
        # Seems slower due to memory
        #hypo_inliers_flag = np.logical_and.reduce(
        #    [xy_inliers_flag, ori_inliers_flag, scale_inliers_flag])
        # this is also slower
        #hypo_inliers_flag = np.logical_and.reduce((xy_inliers_flag,
        #ori_inliers_flag, scale_inliers_flag), out=xy_inliers_flag)
        refined_inliers = np.where(hypo_inliers_flag)[0].astype(INDEX_DTYPE)
        refined_errors = (xy_err, ori_err, scale_err)
    else:
        refined_inliers = np.where(xy_err < xy_thresh_sqrd)[0].astype(INDEX_DTYPE)
        refined_errors = (xy_err, None, None)
    homog_tup1 = (refined_inliers, refined_errors, H)
    return homog_tup1


def test_affine_errors(H, kpts1, kpts2, fm, xy_thresh_sqrd, scale_thresh_sqrd,
                       ori_thresh):
    """
    used for refinement as opposed to initial estimation
    """
    kpts1_m = kpts1.take(fm.T[0], axis=0)
    kpts2_m = kpts2.take(fm.T[1], axis=0)
    invVR1s_m = ktool.get_invVR_mats3x3(kpts1_m)
    xy2_m  = ktool.get_xys(kpts2_m)
    det2_m = ktool.get_sqrd_scales(kpts2_m)
    ori2_m = ktool.get_oris(kpts2_m)
    refined_inliers, refined_errors = _test_hypothesis_inliers(
        H, invVR1s_m, xy2_m, det2_m, ori2_m, xy_thresh_sqrd, scale_thresh_sqrd,
        ori_thresh)
    refined_tup1 = (refined_inliers, refined_errors, H)
    return refined_tup1


def refine_inliers(kpts1, kpts2, fm, aff_inliers, xy_thresh_sqrd,
                   scale_thresh=2.0, ori_thresh=1.57, full_homog_checks=True,
                   refine_method='homog'):
    """
    Given a set of hypothesis inliers, computes a homography and refines inliers
    returned homography maps image1 space into image2 space

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-refine_inliers
        python -m vtool_ibeis.spatial_verification --test-refine_inliers:0
        python -m vtool_ibeis.spatial_verification --test-refine_inliers:1 --show

    Example0:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.demodata as demodata
        >>> import vtool_ibeis.keypoint as ktool
        >>> kpts1, kpts2 = demodata.get_dummy_kpts_pair((100, 100))
        >>> fm = demodata.make_dummy_fm(len(kpts1)).astype(np.int32)
        >>> aff_inliers = np.arange(len(fm))
        >>> xy_thresh_sqrd = .01 * ktool.get_kpts_dlen_sqrd(kpts2)
        >>> homogtup = refine_inliers(kpts1, kpts2, fm, aff_inliers, xy_thresh_sqrd)
        >>> refined_inliers, refined_errors, H = homogtup
        >>> import ubelt as ub
        >>> result = ub.repr2(homogtup, precision=2, nl=True, suppress_small=True, nobr=True)
        >>> print(result)

    Example1:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.spatial_verification import *  # NOQA
        >>> import vtool_ibeis.keypoint as ktool
        >>> kpts1, kpts2, fm, aff_inliers, rchip1, rchip2, xy_thresh_sqrd = testdata_matching_affine_inliers()
        >>> homog_tup1 = refine_inliers(kpts1, kpts2, fm, aff_inliers, xy_thresh_sqrd)
        >>> homog_tup = (homog_tup1[0], homog_tup1[2])
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.draw_sv.show_sv(rchip1, rchip2, kpts1, kpts2, fm, homog_tup=homog_tup)
        >>> ut.show_if_requested()

    """
    H = estimate_refined_transform(kpts1, kpts2, fm, aff_inliers,
                                   refine_method=refine_method)
    if refine_method.endswith('homog'):
        homog_tup1 = test_homog_errors(H, kpts1, kpts2, fm, xy_thresh_sqrd,
                                       scale_thresh, ori_thresh, full_homog_checks)
    #elif refine_method == 'cv2-homog':
    #    homog_tup1 = test_homog_errors(H, kpts1, kpts2, fm, xy_thresh_sqrd,
    #                                   scale_thresh, ori_thresh, full_homog_checks)
    elif refine_method == 'affine':
        homog_tup1 = test_affine_errors(H, kpts1, kpts2, fm, xy_thresh_sqrd,
                                        scale_thresh, ori_thresh)
    return homog_tup1


def get_best_affine_inliers_(kpts1, kpts2, fm, fs, xy_thresh_sqrd,
                             scale_thresh, ori_thresh):
    if HAVE_SVER_C_WRAPPER:
        aff_inliers, aff_errors, Aff = sver_c_wrapper.get_best_affine_inliers_cpp(
            kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh, ori_thresh)
    else:
        if ut.NOT_QUIET:
            print('WARNING: sver has not been compiled')
        aff_inliers, aff_errors, Aff = get_best_affine_inliers(
            kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh, ori_thresh)
    return aff_inliers, aff_errors, Aff


def spatially_verify_kpts(kpts1, kpts2, fm,
                          xy_thresh=.01,
                          scale_thresh=2.0,
                          ori_thresh=TAU / 4.0,
                          dlen_sqrd2=None,
                          min_nInliers=4,
                          match_weights=None,
                          returnAff=False,
                          full_homog_checks=True,
                          refine_method='homog',
                          max_nInliers=5000,
                          ):
    """
    Driver function
    Spatially validates feature matches

    FIXME: there is a non-determenism here

    Returned homography maps image1 space into image2 space.

    Args:
        kpts1 (ndarray[ndim=2]): all keypoints in image 1
        kpts2 (ndarray[ndim=2]): all keypoints in image 2
        fm (ndarray[ndim=2]): matching keypoint indexes [..., (kp1x, kp2x), ...]
        xy_thresh (float): spatial distance threshold under affine transform to
                           be considered a match
        scale_thresh (float):
        ori_thresh (float):
        dlen_sqrd2 (float): diagonal length squared of image/chip 2
        min_nInliers (int): default=4
        returnAff (bool): returns best affine hypothesis as well
        max_nInliers (int): homog is not considered after this threshold

    Returns:
        tuple : (refined_inliers, refined_errors, H, aff_inliers, aff_errors, Aff) if success else None

    CommandLine:
        python -m vtool_ibeis.spatial_verification --test-spatially_verify_kpts --show
        python -m vtool_ibeis.spatial_verification --test-spatially_verify_kpts --show --refine-method='affine'
        python -m vtool_ibeis.spatial_verification --test-spatially_verify_kpts --dpath figures --show --save ~/latex/crall-candidacy-2015/figures/sver_kpts.jpg  # NOQA
        python -m vtool_ibeis.spatial_verification --test-spatially_verify_kpts

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:pyhesaff)
        >>> from vtool_ibeis.spatial_verification import *
        >>> import vtool_ibeis.demodata as demodata
        >>> import vtool_ibeis as vt
        >>> fname1 = ut.get_argval('--fname1', type_=str, default='tsukuba_r')
        >>> fname2 = ut.get_argval('--fname2', type_=str, default='tsukuba_l')
        >>> default_dict = vt.get_extract_features_default_params()
        >>> default_dict['ratio_thresh'] = .625
        >>> kwargs = ut.argparse_dict(default_dict)
        >>> (kpts1, kpts2, fm, fs, rchip1, rchip2) = demodata.testdata_ratio_matches(fname1, fname2, **kwargs)
        >>> xy_thresh = .01
        >>> dlen_sqrd2 = 447271.015
        >>> ori_thresh = 1.57
        >>> min_nInliers = 4
        >>> returnAff = True
        >>> scale_thresh = 2.0
        >>> match_weights = np.ones(len(fm), dtype=np.float64)
        >>> refine_method = ut.get_argval('--refine-method', default='homog')
        >>> svtup = spatially_verify_kpts(kpts1, kpts2, fm, xy_thresh,
        >>>                               scale_thresh, ori_thresh, dlen_sqrd2,
        >>>                               min_nInliers, match_weights, returnAff,
        >>>                               refine_method=refine_method)
        >>> assert svtup is not None and len(svtup) == 6, 'sver failed'
        >>> refined_inliers, refined_errors, H = svtup[0:3]
        >>> aff_inliers, aff_errors, Aff = svtup[3:6]
        >>> #print('aff_errors = %r' % (aff_errors,))
        >>> print('aff_inliers = %r' % (aff_inliers,))
        >>> print('refined_inliers = %r' % (refined_inliers,))
        >>> #print('refined_errors = %r' % (refined_errors,))
        >>> import ubelt as ub
        >>> result = ut.list_type_profile(svtup, with_dtype=False)
        >>> #result = ub.repr2(svtup, precision=3)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> homog_tup = (refined_inliers, H)
        >>> aff_tup = (aff_inliers, Aff)
        >>> pt.draw_sv.show_sv(rchip1, rchip2, kpts1, kpts2, fm, aff_tup=aff_tup, homog_tup=homog_tup, refine_method=refine_method)
        >>> pt.show_if_requested()
    """
    if len(fm) == 0:
        if VERBOSE_SVER:
            print('[sver] Cannot verify with no matches')
        svtup = None
        return svtup
    # Cast keypoints to float64 to avoid numerical issues
    kpts1 = kpts1.astype(np.float64, casting='same_kind', copy=False)
    kpts2 = kpts2.astype(np.float64, casting='same_kind', copy=False)
    #kpts1 = kpts1.astype(np.float64)
    #kpts2 = kpts2.astype(np.float64)
    assert match_weights is not None, 'provide at least ones please for match_weights'
    fs = match_weights
    # Get diagonal length if not provided
    if dlen_sqrd2 is None:
        kpts2_m = kpts2.take(fm.T[1], axis=0)
        dlen_sqrd2 = ktool.get_kpts_dlen_sqrd(kpts2_m)
    # Determine the best hypothesis transformation and get its inliers
    xy_thresh_sqrd = dlen_sqrd2 * xy_thresh
    aff_inliers, aff_errors, Aff = get_best_affine_inliers_(
        kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh, ori_thresh)
    #print(aff_inliers)

    # Return if there are not enough inliers to compute homography
    if len(aff_inliers) < min_nInliers:
        # Test user defined param
        if VERBOSE_SVER:
            print('[sver] Failed spatial verification len(aff_inliers) = %r' %
                  (len(aff_inliers),))
        svtup = None
        return svtup
    if ((refine_method.endswith('homog') and len(aff_inliers) < 7) or
         len(aff_inliers) < 4):
        # Test fundamental param
        # need to have 4 or more inliers to comopute an affine
        # and need at least 7 to compute a homography
        if VERBOSE_SVER:
            print('[sver] Failed spatial verification len(aff_inliers) = %r' %
                  (len(aff_inliers),))
        svtup = None
        return svtup

    if len(aff_inliers) >= max_nInliers:
        # If there are a very large number of affine inliers, then the affine
        # matrix is probably good enough.
        svtup = (aff_inliers, aff_errors, Aff, aff_inliers, aff_errors, Aff)
        return svtup

    # Refine inliers using a projective transformation (homography)
    try:
        refined_inliers, refined_errors, H = refine_inliers(
            kpts1, kpts2, fm, aff_inliers, xy_thresh_sqrd, scale_thresh,
            ori_thresh, full_homog_checks, refine_method=refine_method)
        #print(refined_inliers)
    except npl.LinAlgError as ex:
        if ut.VERYVERBOSE and ut.SUPER_STRICT:
            ut.printex(ex, 'numeric error in homog estimation.', iswarning=True)
        return None
    except ValueError as ex:
        if ut.VERYVERBOSE and ut.SUPER_STRICT:
            ut.printex(ex, 'error cv2 in homog estimation.', iswarning=True)
        return None
    except IndexError:
        raise
    except Exception as ex:
        # There is a weird error that starts with MemoryError and ends up
        # makeing len(h) = 6.
        ut.printex(ex, 'Unknown error in homog estimation.',
                      keys=['kpts1', 'kpts2',  'fm', 'fm.shape', 'kpts1.shape',
                            (len, 'aff_inliers'),
                            'kpts2.shape', 'xy_thresh', 'scale_thresh',
                            'dlen_sqrd2', 'min_nInliers'])
        if ut.SUPER_STRICT:
            print('SUPER_STRICT is on. Reraising')
            raise
        return None
    if VERBOSE_SVER:
        print('[sver] Succesfully finished spatial verification.')
    if returnAff:
        svtup = (refined_inliers, refined_errors, H, aff_inliers, aff_errors, Aff)
        return svtup
    else:
        svtup = (refined_inliers, refined_errors, H, None, None, None)
        return svtup


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.spatial_verification
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
# LICENCE Apache 2 or whatever

FIXME: monotization functions need more hueristics
"""
import numpy as np
import utool as ut  # NOQA
import ubelt as ub  # NOQA


TAU = np.pi * 2  # References: tauday.com

eps = 1E-9


def interpolate_nans(arr):
    r"""
    replaces nans with interpolated values or 0

    Args:
        arr (ndarray):

    Returns:
        ndarray: new_arr

    CommandLine:
        python -m vtool_ibeis.util_math --exec-interpolate_nans

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> arr = np.array([np.nan, np.nan, np.nan, np.nan])
        >>> new_arr = interpolate_nans(arr)
        >>> result = ('new_arr = %s' % (str(new_arr),))
        >>> print(result)
        new_arr = [ 0.  0.  0.  0.]

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> arr = np.array([np.nan, 1, np.nan, np.nan, np.nan, np.nan, 10, np.nan, 5])
        >>> new_arr = interpolate_nans(arr)
        >>> result = ('new_arr = %s' % (str(new_arr),))
        >>> print(result)
        new_arr = [  1.    1.    2.8   4.6   6.4   8.2  10.    7.5   5. ]
    """
    import vtool_ibeis as vt
    new_arr = arr.copy()
    nan_idxs = np.where(np.isnan(arr))[0]
    consecutive_groups = vt.group_consecutive(nan_idxs)
    last_index = len(arr) - 1
    for group in consecutive_groups:
        min_ = group.min()
        max_ = group.max()
        if min_ == 0 and max_ == last_index:
            upper = lower = 0
        else:
            if min_ != 0:
                lower = arr[min_ - 1]
            if max_ != last_index:
                upper = arr[max_ + 1]
            if min_ == 0:
                lower = upper
            if max_ == last_index:
                upper = lower
        new_arr[min_:max_ + 1] = np.linspace(lower, upper, len(group) + 2)[1:-1]
    return new_arr


def ensure_monotone_strictly_increasing(arr_, left_endpoint=None, right_endpoint=None, zerohack=False, onehack=False, newmode=True):
    """

    Args:
        arr_ (ndarray): sequence to monotonize
        zerohack (bool): default False, if True sets the first element to be zero and linearlly interpolates to the first nonzero item
        onehack (bool):  default False, if True one will not be in the resulting array (replaced with number very close to one)

    References:
        http://mathoverflow.net/questions/17464/making-a-non-monotone-function-monotone
        http://stackoverflow.com/questions/28563711/make-a-numpy-array-monotonic-without-a-python-loop
        https://en.wikipedia.org/wiki/Isotonic_regression
        http://scikit-learn.org/stable/auto_examples/plot_isotonic_regression.html

    CommandLine:
        python -m vtool_ibeis.util_math --test-ensure_monotone_strictly_increasing --show
        python -m vtool_ibeis.util_math --test-ensure_monotone_strictly_increasing --show --offset=0

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *   # NOQA
        >>> import numpy as np
        >>> arr_ = np.array([0.4, 0.4, 0.4, 0.5, 0.6, 0.6, 0.6, 0.7, 0.9, 0.9, 0.91, 0.92, 1.0, 1.0])
        >>> arr = ensure_monotone_strictly_increasing(arr_)
        >>> assert strictly_increasing(arr), 'ensure strict monotonic failed1'

    Example2:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> left_endpoint = None
        >>> rng = np.random.RandomState(0)
        >>> right_endpoint = None
        >>> domain = np.arange(100)
        >>> offset = ut.get_argval('--offset', type_=float, default=2.3)
        >>> arr_ = np.sin(np.pi * (domain / 100) - offset) + (rng.rand(len(domain)) - .5) * .1 + 1.2
        >>> #arr_ = vt.demodata.testdata_nonmonotonic()
        >>> #domain = np.arange(len(arr_))
        >>> arr = ensure_monotone_strictly_increasing(arr_, left_endpoint, right_endpoint)
        >>> result = str(arr)
        >>> print(result)
        >>> print('arr = %r' % (arr,))
        >>> print('arr = %r' % (np.diff(arr),))
        >>> assert non_decreasing(arr), 'ensure nondecreasing failed2'
        >>> assert strictly_increasing(arr), 'ensure strict monotonic failed2'
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot2(domain, arr_, 'r-', fnum=1, pnum=(3, 1, 1), title='before', equal_aspect=False)
        >>> arr2 = ensure_monotone_increasing(arr_)
        >>> pt.plot2(domain, arr, 'b-', fnum=1, pnum=(3, 1, 2), equal_aspect=False)
        >>> pt.plot2(domain, arr2, 'r-', fnum=1, pnum=(3, 1, 2), title='after monotonization (decreasing)', equal_aspect=False)
        >>> pt.plot2(domain, arr, 'r-', fnum=1, pnum=(3, 1, 3), title='after monotonization (strictly decreasing)', equal_aspect=False)
        >>> ut.show_if_requested()
    """
    arr = ensure_monotone_increasing(arr_, newmode=newmode)
    if zerohack:
        left_endpoint = 0.0
    if onehack:
        right_endpoint = 1.0
    #print('arr_in = %r' % (arr,))
    #print('right_endpoint = %r' % (right_endpoint,))
    #print('left_endpoint = %r' % (left_endpoint,))
    arr = breakup_equal_streak(arr, left_endpoint, right_endpoint)
    #print('arr_out = %r' % (arr,))
    return arr


def ensure_monotone_strictly_decreasing(arr_, left_endpoint=None, right_endpoint=None):
    """

    Args:
        arr_ (ndarray):
        left_endpoint (None):
        right_endpoint (None):

    Returns:
        ndarray: arr

    CommandLine:
        python -m vtool_ibeis.util_math --test-ensure_monotone_strictly_decreasing --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> domain = np.arange(100)
        >>> rng = np.random.RandomState(0)
        >>> arr_ = np.sin(np.pi * (domain / 75) + 1.3) + (rng.rand(len(domain)) - .5) * .05 + 1.0
        >>> #arr_ = vt.demodata.testdata_nonmonotonic()
        >>> #domain = np.arange(len(arr_))
        >>> left_endpoint = 2.5
        >>> right_endpoint = 0.25
        >>> arr = ensure_monotone_strictly_decreasing(arr_, left_endpoint, right_endpoint)
        >>> result = str(arr)
        >>> print(result)
        >>> assert strictly_decreasing(arr), 'ensure strict monotonic failed'
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot2(domain, arr_, 'r-', fnum=1, pnum=(3, 1, 1), title='before', equal_aspect=False)
        >>> arr2 = ensure_monotone_decreasing(arr_)
        >>> pt.plot2(domain, arr, 'b-', fnum=1, pnum=(3, 1, 2), equal_aspect=False)
        >>> pt.plot2(domain, arr2, 'r-', fnum=1, pnum=(3, 1, 2), title='after monotonization (decreasing)', equal_aspect=False)
        >>> pt.plot2(domain, arr, 'r-', fnum=1, pnum=(3, 1, 3), title='after monotonization (strictly decreasing)', equal_aspect=False)
        >>> ut.show_if_requested()
    """
    #raise NotImplementedError('unfinished')
    arr = ensure_monotone_decreasing(arr_)
    # FIXME: doesn't work yet I don't think
    arr = breakup_equal_streak(arr, left_endpoint, right_endpoint)
    return arr


def breakup_equal_streak(arr_in, left_endpoint=None, right_endpoint=None):
    """
    Breaks up streaks of equal values by interpolating between the next lowest and next highest value

    Args:
        arr_in (?):
        left_endpoint (None): (default = None)
        right_endpoint (None): (default = None)

    Returns:
        ndarray: arr -

    CommandLine:
        python -m vtool_ibeis.util_math --exec-breakup_equal_streak
        python -m vtool_ibeis.util_math --test-ensure_monotone_strictly_increasing --show --offset=0

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> arr_in = np.array([0, 0, 1, 1, 2, 2], dtype=np.float32)
        >>> arr_in = np.array([ 1.20488135,  1.2529297 ,  1.27306686,  1.29859663,
        >>>    1.31769871, 1.37102388,  1.38114004,  1.45732054,  1.48119571,  1.48119571,
        >>>     1.5381895 ,  1.54162741,  1.57492901,  1.61129523,  1.61129523,
        >>>     1.61270343,  1.63377551,  1.7423034 ,  1.76364247,  1.79908459,
        >>>     1.83564709,  1.83819742,  1.83819742,  1.86786967,  1.86786967,
        >>>     1.90720142,  1.90720142,  1.92293973,  1.92293973, ]) / 2
        >>> left_endpoint = 0
        >>> right_endpoint = 1.0
        >>> arr = breakup_equal_streak(arr_in, left_endpoint, right_endpoint)
        >>> assert strictly_increasing(arr)
        >>> result = ('arr = %s' % (str(arr),))
        >>> print(result)
    """
    #assert non_decreasing(arr), 'ensure monotonic failed'
    arr = arr_in.copy()

    # Find maxish and minish before adjusting
    if right_endpoint is not None:
        # Dont be so confident
        # Set the rightmost value to be almost ``right_endpoint``
        range_ = np.abs(arr_in[0] - arr_in[-1])
        if arr_in[0] < right_endpoint:
            # increasing arr
            almost_right = right_endpoint - (range_ * .001)
            # The second highest value in arr, or close enough
            maxish = min(almost_right, arr[arr < right_endpoint].max())
            newmax = (right_endpoint * .9 + maxish * .1)
        else:
            # decreasing arr
            almost_right = right_endpoint + (range_ * .001)
            # The second lowest value in arr, or close enough
            minish = max(almost_right, arr[arr > right_endpoint].min())
            #newmin = (right_endpoint + minish) / 2.0
            newmin = (right_endpoint * .9 + minish * .1)

    size = len(arr)
    is_same = np.abs(np.diff(arr)) < 1E-8
    #is_same = np.diff(arr) == 0
    #index_list = np.nonzero(np.diff(arr) == 0)[0]
    index_list = np.nonzero(is_same)[0]
    if len(index_list) == 0:
        # If there are no consecutive numbers then arr must be strictly
        # increasing
        return arr

    consecutive_groups = group_consecutive(index_list)
    index_groups = [np.array(group.tolist() + [group.max() + 1]) for group in consecutive_groups]
    # Nope this is right
    # Hack because sometimes things arent't grouped correctly
    # items in index groups are consectuive and breaking things
    # arr[ut.flatten(index_groups)]
    #index_groups2 = []
    #for group in index_groups:
    #    if len(index_groups2) == 0:
    #        index_groups2.append(group)
    #    elif index_groups2[-1][-1] + 1 == group[0]:
    #        print('group = %r' % (group,))
    #        # JOIN CASE
    #    else:
    #        index_groups2.append(group)

    runlen_list = [len(group) for group in index_groups]
    # Handle ending corner case

    #isend_list = [(group[-1] + 1) < size for group in index_groups]
    # Error? Should this be less?
    #isend_list = [(group[-1] + 1) < size for group in index_groups]
    isend_list = [(group[-1] + 1) >= size for group in index_groups]
    isstart_list = [group[0] == 0 for group in index_groups]

    min_vals = [
        arr[group[0]]
        if isstart else
        (.49 * arr[group[0] - 1] + .51 * arr[group[0]])  # value between previous and this one (bumped to right)
        for group, isstart in zip(index_groups, isstart_list)
    ]

    max_vals = [
        arr[group[-1]]  # Max value is the value of the previous group?
        if isend else
        (.49 * arr[group[-1] + 1] + .51 * arr[group[-1]])  # value between next and this one (bumped to right)
        for group, isend in zip(index_groups, isend_list)
    ]

    #import vtool_ibeis as vt
    #vt.apply_grouping(arr, index_groups)
    #np.vstack((min_vals, max_vals)).T

    fill_list = [np.linspace(min_, max_, len_, endpoint=not isend)
                 for min_, max_, len_, isend in zip(min_vals, max_vals, runlen_list, isend_list)]

    for group, fill in zip(index_groups, fill_list):
        arr[group[0]:group[-1] + 1] = fill

    if left_endpoint is not None and len(index_groups) > 0:
        # Set the leftmost value to be exactly ``left_endpoint``
        group_ = index_groups[0]
        if group_[0] == 0:
            group_0_slice = slice(group_[0], group_[-1] + 1)
            arr[group_0_slice] = np.linspace(left_endpoint, arr[group_[-1]], len(group_))
        else:
            arr[0] = left_endpoint
    if right_endpoint is not None:
        # Dont be so confident
        # Set the rightmost value to be almost ``right_endpoint``
        if arr_in[0] < right_endpoint:
            # increasing arr
            if len(isend_list) > 0 and isend_list[-1]:
                # adjust for adjustments
                maxish = min(min_vals[-1], maxish)
            arr[arr >= maxish] = np.linspace(maxish, newmax, sum(arr >= maxish))
        else:
            if len(isstart_list) > 0 and isstart_list[0]:
                minish = max(max_vals[0], minish)
            # decreasing arr
            arr[arr <= minish] = np.linspace(minish, newmin, sum(arr <= minish))
    return arr


def group_consecutive(arr):
    """
    Returns lists of consecutive values

    References:
        http://stackoverflow.com/questions/7352684/how-to-find-the-groups-of-consecutive-elements-from-an-array-in-numpy

    Args:
        arr (ndarray): must be integral and unique

    Returns:
        ndarray: arr -

    CommandLine:
        python -m vtool_ibeis.util_math --exec-group_consecutive

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> arr = np.array([1, 2, 3, 5, 6, 7, 8, 9, 10, 15, 99, 100, 101])
        >>> groups = group_consecutive(arr)
        >>> result = ('groups = %s' % (str(groups),))
        >>> print(result)
        groups = [array([1, 2, 3]), array([ 5,  6,  7,  8,  9, 10]), array([15]), array([ 99, 100, 101])]
    """
    #is_nonconsec = np.abs(np.diff(arr)) < 1E-2
    #split_indicies = np.nonzero(is_nonconsec)[0] + 1
    split_indicies = np.nonzero(np.diff(arr) != 1)[0] + 1
    groups = np.array_split(arr, split_indicies)
    return groups
    #return np.array_split(arr, np.where(np.diff(arr) != 1)[0] + 1)


def strictly_increasing(L):
    """
    References:
        http://stackoverflow.com/questions/4983258/python-how-to-check-list-monotonicity
    """
    return all(x < y for x, y in zip(L, L[1:]))


def strictly_decreasing(L):
    """
    References:
        http://stackoverflow.com/questions/4983258/python-how-to-check-list-monotonicity
    """
    return all(x > y for x, y in zip(L, L[1:]))


def non_increasing(L):
    """
    References:
        http://stackoverflow.com/questions/4983258/python-how-to-check-list-monotonicity
    """
    return all(x >= y for x, y in zip(L, L[1:]))


def non_decreasing(L):
    """
    References:
        http://stackoverflow.com/questions/4983258/python-how-to-check-list-monotonicity
    """
    return all(x <= y for x, y in zip(L, L[1:]))


def ensure_monotone_increasing(arr_, fromright=True, fromleft=True, newmode=True):
    r"""
    Args:
        arr_ (ndarray):

    Returns:
        ndarray: arr

    CommandLine:
        python -m vtool_ibeis.util_math --test-ensure_monotone_increasing --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> rng = np.random.RandomState(0)
        >>> size_ = 100
        >>> domain = np.arange(size_)
        >>> offset = float(ub.argval('--offset', default=2.3))
        >>> arr_ = np.sin(np.pi * (domain / 100) - offset) + (rng.rand(len(domain)) - .5) * .1
        >>> arr = ensure_monotone_increasing(arr_, fromleft=False, fromright=True)
        >>> result = str(arr)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot2(domain, arr_, 'r-', fnum=1, pnum=(2, 1, 1), title='before', equal_aspect=False)
        >>> pt.plot2(domain, arr, 'r-', fnum=1, pnum=(2, 1, 2), title='after monotonization (increasing)', equal_aspect=False)
        >>> ut.show_if_requested()
    """
    if newmode:
        from sklearn.isotonic import IsotonicRegression
        ir = IsotonicRegression()
        arr = ir.fit_transform(np.arange(len(arr_)), arr_)
    else:
        arr = arr_.copy()
        size = len(arr)
        # Ensure increasing from right
        if fromright:
            for lx in range(1, size):
                rx = (size - lx - 1)
                if arr[rx] > arr[rx + 1]:
                    arr[rx] = arr[rx + 1]
        if fromleft:
            # ensure increasing from left
            for lx in range(0, size - 1):
                if arr[lx] > arr[lx + 1]:
                    arr[lx + 1] = arr[lx]
    return arr


def ensure_monotone_decreasing(arr_, fromleft=True, fromright=True):
    r"""
    Args:
        arr_ (ndarray):

    Returns:
        ndarray: arr

    CommandLine:
        python -m vtool_ibeis.util_math --test-ensure_monotone_decreasing --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> rng = np.random.RandomState(0)
        >>> size_ = 100
        >>> domain = np.arange(size_)
        >>> arr_ = np.sin(np.pi * (domain / 100) ) + (rng.rand(len(domain)) - .5) * .1
        >>> arr = ensure_monotone_decreasing(arr_, fromright=True, fromleft=True)
        >>> result = str(arr)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot2(domain, arr_, 'r-', fnum=1, pnum=(2, 1, 1), title='before', equal_aspect=False)
        >>> pt.plot2(domain, arr, 'r-', fnum=1, pnum=(2, 1, 2), title='after monotonization (decreasing)', equal_aspect=False)
        >>> ut.show_if_requested()
    """
    arr = arr_.copy()
    size = len(arr)
    if fromright:
        # Ensure decreasing from right
        for lx in range(1, size):
            rx = (size - lx - 1)
            if arr[rx] < arr[rx + 1]:
                arr[rx] = arr[rx + 1]
    if fromleft:
        # ensure increasing from left
        for lx in range(0, size - 1):
            if arr[lx] < arr[lx + 1]:
                arr[lx + 1] = arr[lx]
    return arr


def test_language_modulus():
    """
    References:
        http://en.wikipedia.org/wiki/Modulo_operation
    """
    import math
    import utool as ut  # NOQA
    TAU = math.pi * 2
    num_list = [-8, -1, 0, 1, 2, 6, 7, 29]
    modop_result_list = []
    fmod_result_list = []
    for num in num_list:
        num = float(num)
        modop_result_list.append(num % TAU)
        fmod_result_list.append(math.fmod(num, TAU))
    table = ut.make_csv_table([num_list, modop_result_list, fmod_result_list],
                              ['num',  'modop', 'fmod'], 'mods', [float, float, float] )
    print(table)


def iceil(num, dtype=np.int32):
    """ Integer ceiling. (because numpy doesn't seem to have it!)

    Args:
        num (ndarray or scalar):

    Returns:
        ndarray or scalar:

    CommandLine:
        python -m vtool_ibeis.util_math --test-iceil

    Setup:
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> import vtool_ibeis as vt

    Example0:
        >>> # ENABLE_DOCTEST
        >>> num = 1.5
        >>> result = repr(vt.iceil(num))
        >>> print(result)
        2

    Example1:
        >>> # ENABLE_DOCTEST
        >>> num = [1.5, 2.9]
        >>> result = ub.repr2(vt.iceil(num), with_dtype=True)
        >>> print(result)
        np.array([2, 3], dtype=np.int32)
    """
    return np.ceil(num).astype(dtype)


def iround(num, dtype=np.int32):
    """ Integer round. (because numpy doesn't seem to have it!) """
    return np.round(num).astype(dtype)


def gauss_func1d(x, mu=0.0, sigma=1.0):
    r"""
    Args:
        x (?):
        mu (float):
        sigma (float):

    CommandLine:
        python -m vtool_ibeis.util_math --test-gauss_func1d

    CommandLine:
        python -m vtool_ibeis.util_math --test-gauss_func1d --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> # build test data
        >>> x = np.array([-2, -1, -.5, 0, .5, 1, 2])
        >>> mu = 0.0
        >>> sigma = 1.0
        >>> # execute function
        >>> gaussval = gauss_func1d(x, mu, sigma)
        >>> # verify results
        >>> result = np.array_repr(gaussval, precision=2)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot(x, gaussval)
        >>> ut.show_if_requested()
        array([ 0.05,  0.24,  0.35,  0.4 ,  0.35,  0.24,  0.05])
    """
    coeff = np.reciprocal(sigma * np.sqrt(TAU))
    exponent_expr_numer = np.power(np.subtract(x, mu), 2)
    exponent_expr_denom = (-2 * (sigma ** 2))
    exponent_expr = np.divide(exponent_expr_numer, exponent_expr_denom)
    gaussval = coeff * np.exp(exponent_expr)
    return gaussval


def gauss_func1d_unnormalized(x, sigma=1.0):
    """
    faster version of gauss_func1d with no normalization. So the maximum point
    will have a value of 1.0


    CommandLine:
        python -m vtool_ibeis.util_math --test-gauss_func1d_unnormalized --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> # build test data
        >>> x = np.array([-2, -1, -.5, 0, .5, 1, 2])
        >>> sigma = 1.0
        >>> # execute function
        >>> gaussval = gauss_func1d_unnormalized(x, sigma)
        >>> # verify results
        >>> result = np.array_repr(gaussval, precision=2)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot(x, gaussval)
        >>> ut.show_if_requested()
        array([ 0.05,  0.24,  0.35,  0.4 ,  0.35,  0.24,  0.05])
    """
    exponent_expr_denom = (-2 * (sigma ** 2))
    tmp = exponent_expr_numer = np.power(x, 2.0)
    exponent_expr = np.divide(exponent_expr_numer, exponent_expr_denom, out=tmp)
    gaussval = np.exp(exponent_expr, out=tmp)
    return gaussval


def logistic_01(x):
    r"""
    Args:
        x (?):

    CommandLine:
        python -m vtool_ibeis.util_math --exec-logistic_01 --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_math import *  # NOQA
        >>> x = np.linspace(0, 1)
        >>> y = logistic_01(x)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot(x, y)
        >>> ut.show_if_requested()
    """
    from scipy.special import expit
    y = expit(((x * 2) - 1.0) * 6)
    return y
    # return L / (1 + np.exp(-k * (x - x0)))


def logit(x):
    from scipy.special import logit
    return logit(x)


# def relu_01(x, p):
#     return np.minimum(0, (x - p) / (1 - p))


def beaton_tukey_loss(u, a=1):
    """
    CommandLine:
        python -m plottool_ibeis.draw_func2 --exec-plot_func --show --range=-8,8 --func=vt.beaton_tukey_weight,vt.beaton_tukey_loss

    References:
        Steward_Robust%20parameter%20estimation%20in%20computer%20vision.pdf
    """
    result = np.empty(u.shape, dtype=u.dtype)
    is_case1 = np.abs(u) <= a
    u1 = u[is_case1]
    result[is_case1] = ((a ** 2) / 6) * (1 - (1 - (u1 / a) ** 2) ** 3)
    result[~is_case1] = (a ** 2 / 6)
    return result


def beaton_tukey_weight(u, a=1):
    """
    CommandLine:
        python -m plottool_ibeis.draw_func2 --exec-plot_func --show --range=-8,8 --func=vt.beaton_tukey_weight

    References:
        Steward_Robust%20parameter%20estimation%20in%20computer%20vision.pdf
    """
    result = np.empty(u.shape, dtype=u.dtype)
    is_case1 = np.abs(u) <= a
    u1 = u[is_case1]
    result[is_case1] = u1 * (1 - (u1 / a) ** 2) ** 2
    result[~is_case1] = 0
    return result


def gauss_parzen_est(dist, L=1, sigma=.38):
    """
    python -m plottool_ibeis.draw_func2 --exec-plot_func --show --range=-.2,.2 --func=vt.gauss_parzen_est
    python -m plottool_ibeis.draw_func2 --exec-plot_func --show --range=0,1 --func=vt.gauss_parzen_est
    """
    tau = np.pi * 2
    const_term = np.log(L * sigma * np.sqrt(tau))
    return np.exp((-dist / (2 * sigma ** 2)) - const_term)


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.util_math
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



from __future__ import absolute_import, division, print_function
# import utool

# def test_akmeans(full_test=False, plot_test=False, num_pca_dims=2, data_dim=2,
#                  nump=1000):
#     import numpy as np
#     from vtool_ibeis import clustering
#     nump = nump
#     dims = data_dim  # 128
#     dtype = np.uint8
#     print('Make %d random %d-dimensional %s points.' % (nump, dims, dtype))
#     # Seed for a determenistic test
#     np.random.seed(42)
#     data = np.array(np.random.randint(0, 255, (nump, dims)), dtype=dtype)

#     num_clusters = 10
#     max_iters = 2
#     ave_unchanged_thresh = 0
#     ave_unchanged_iterwin = 10
#     flann_params = {}

#     cache_dir = utool.get_app_resource_dir('vtool_ibeis', 'test_cache')
#     utool.ensuredir(cache_dir)

#     # Test precomputing
#     dx2_label, centers = clustering.precompute_akmeans(data, num_clusters,
#                                                        max_iters=max_iters,
#                                                        cache_dir=cache_dir)
#     # internal names
#     datax2_clusterx, centroids = dx2_label, centers

#     if plot_test:
#         clustering.plot_clusters(data, datax2_clusterx, centroids, num_pca_dims=num_pca_dims)

#     assert centers.shape == (num_clusters, dims), 'sanity check'
#     assert dx2_label.shape == (nump,), 'sanity check'

#     # Test regular computing
#     if full_test:
#         dx2_label, centers = clustering.akmeans(data, num_clusters, max_iters=max_iters)
#         assert centers.shape == (num_clusters, dims), 'sanity check'
#         assert dx2_label.shape == (nump,), 'sanity check'

#     if False:
#         # other test (development)
#         from vtool_ibeis._pyflann_backend import pyflann
#         flann_lib_inst = pyflann.flann
#         flann_class_inst = pyflann.FLANN()
#         flann_class_inst.build_index(data)
#     return locals()


# if __name__ == '__main__':
#     testkw = {
#         'plot_test': utool.get_argflag('--plot-test'),
#         'full_test': utool.get_argflag('--full-test'),
#         'num_pca_dims': utool.get_argval('--num-pca-dims', type_=int, default=2),
#         'data_dim': utool.get_argval('--data-dim', type_=int, default=2),
#         'nump': utool.get_argval('--nump', type_=int, default=2000),
#     }
#     test_locals = utool.run_test(test_akmeans, **testkw)
#     exec(utool.execstr_dict(test_locals, 'test_locals'))
#     if testkw['plot_test']:
#         from plottool_ibeis import draw_func2 as df2
#         exec(df2.present())
#     else:
#         exec(utool.ipython_execstr())
